import logging
import time
import os
from typing import List
import numpy as np
from collections import defaultdict
from multiprocessing.pool import Pool
from pymilvus import Collection, DataType, MilvusClient, connections

from langchain.prompts import PromptTemplate
from langchain_core.output_parsers import StrOutputParser
from langchain_openai.chat_models import ChatOpenAI

from arklex.env.prompts import load_prompts
from arklex.utils.mysql import mysql_pool
from arklex.utils.model_config import MODEL
from arklex.env.prompts import load_prompts
from arklex.utils.graph_state import MessageState
from arklex.env.tools.RAG.retrievers.retriever_document import RetrieverDocument, RetrieverDocumentType, RetrieverResult, embed, embed_retriever_document

EMBED_DIMENSION = 1536
MAX_TEXT_LENGTH = 65535
CHUNK_NEIGHBOURS = 3

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class RetrieveEngine():
    @staticmethod
    def milvus_retrieve(state: MessageState):
        # get the input message
        user_message = state['user_message']

        # Search for the relevant documents
        milvus_retriever = MilvusRetrieverExecutor(state["bot_config"])
        retrieved_text, retriever_params = milvus_retriever.retrieve(user_message.history)

        state["message_flow"] = retrieved_text
        state["metadata"]["tool_response"] = retriever_params
        return state

class MilvusRetriever:
    def __enter__(self):
        self.uri = os.getenv("MILVUS_URI", "")
        self.token = os.getenv("MILVUS_TOKEN", "")
        self.client = MilvusClient(uri=self.uri, token=self.token)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.client.close()

    def get_bot_uid(self, bot_id: str, version: str):
        return f"{bot_id}__{version}"

    def create_collection_with_partition_key(self, collection_name: str):
        schema = MilvusClient.create_schema(
            auto_id=False,
            enable_dynamic_field=True,
            partition_key_field="bot_uid",
            num_partitions=16,
        )
        schema.add_field(
            field_name="id", datatype=DataType.VARCHAR, is_primary=True, max_length=100
        )
        schema.add_field(
            field_name="qa_doc_id", datatype=DataType.VARCHAR, max_length=40
        )
        schema.add_field(
            field_name="bot_uid", datatype=DataType.VARCHAR, max_length=100
        )
        schema.add_field(field_name="chunk_id", datatype=DataType.INT32)
        schema.add_field(
            field_name="qa_doc_type", datatype=DataType.VARCHAR, max_length=10
        )
        schema.add_field(field_name="metadata", datatype=DataType.JSON)
        schema.add_field(
            field_name="text", datatype=DataType.VARCHAR, max_length=MAX_TEXT_LENGTH
        )
        schema.add_field(
            field_name="embedding", datatype=DataType.FLOAT_VECTOR, dim=EMBED_DIMENSION
        )
        schema.add_field(field_name="timestamp", datatype=DataType.INT64)
        # schema.add_field(field_name="num_tokens", datatype=DataType.INT64)
        index_params = self.client.prepare_index_params()
        index_params.add_index(field_name="id")
        index_params.add_index(field_name="qa_doc_id")
        index_params.add_index(field_name="bot_uid")
        index_params.add_index(
            field_name="embedding", index_type="FLAT", metric_type="L2"
        )

        self.client.create_collection(
            collection_name=collection_name, schema=schema, index_params=index_params
        )

    def delete_documents_by_qa_doc_id(self, collection_name: str, qa_doc_id: str):
        logger.info(
            f"Deleting vector db documents by qa_doc_id: {qa_doc_id} from collection: {collection_name}"
        )
        res = self.client.delete(
            collection_name=collection_name, filter=f"qa_doc_id=='{qa_doc_id}'"
        )
        return res
    
    def add_documents_dicts(
        self, documents: List[dict], collection_name: str, upsert: bool = False
    ):
        logger.info(f"Celery sub task for adding {len(documents)} documents to collection: {collection_name}.")
        retriever_documents = [RetrieverDocument.from_dict(doc) for doc in documents]
        documents_to_insert = []

        if not upsert:
            # check if the document already exists in the collection
            for doc in retriever_documents:
                res = self.client.get(collection_name=collection_name, ids=doc.id)
                if len(res) == 0:
                    documents_to_insert.append(doc)
            logger.info(f"Exisiting documents: {len(documents_to_insert)}")
        else:
            documents_to_insert = retriever_documents

        res = []
        for doc in documents_to_insert:
            data = doc.to_milvus_schema_dict_and_embed()
            try:
                res.append(
                    self.client.upsert(collection_name=collection_name, data=[data])
                )
            except Exception as e:
                logger.error(f"Error adding document id: {data['id']} error: {e}")
                raise e
        return res
    
    def add_documents_parallel(
        self, collection_name: str, bot_id: str, version: str, documents: List[RetrieverDocument], process_pool: Pool, upsert: bool = False
    ):
        logger.info(
            f"Adding {len(documents)} vector db documents to collection '{collection_name}' for bot_id: {bot_id} version: {version}"
        )
        if not self.client.has_collection(collection_name):
            logger.info(f"No collection found hence creating collection: {collection_name}")
            self.create_collection_with_partition_key(collection_name)

        documents_to_insert = []

        if not upsert:
            # check if the document already exists in the collection
            for doc in documents:
                res = self.client.get(collection_name=collection_name, ids=doc.id)
                if len(res) == 0:
                    documents_to_insert.append(doc)
            logger.info(f"Exisiting documents: {len(documents_to_insert)}")
        else:
            documents_to_insert = documents

        res = []
        # process 100 documents at a time
        count = 0
        for i in range(0, len(documents_to_insert), 100):
            batch_docs = documents_to_insert[i:i+100]
            embedded_batch_docs = process_pool.map(embed_retriever_document, batch_docs)

            res.extend(
                self.client.upsert(collection_name=collection_name, data=embedded_batch_docs)
            )
            count += len(batch_docs)
            logger.info(f"Added {count}/{len(documents_to_insert)} docs")

        return res
        
    def add_documents(
        self, collection_name: str, bot_id: str, version: str, documents: List[RetrieverDocument], upsert: bool = False
    ):
        logger.info(
            f"Adding {len(documents)} vector db documents to collection {collection_name} for bot_id: {bot_id} version: {version}"
        )

        if not self.client.has_collection(collection_name):
            self.create_collection_with_partition_key(collection_name)

        documents_to_insert = []

        if not upsert:
            # check if the document already exists in the collection
            for doc in documents:
                res = self.client.get(collection_name=collection_name, ids=doc.id)
                if len(res) == 0:
                    documents_to_insert.append(doc)
            logger.info(f"Exisiting documents: {len(documents_to_insert)}")
        else:
            documents_to_insert = documents

        res = []
        for doc in documents_to_insert:
            data = doc.to_milvus_schema_dict_and_embed()
            try:
                res.append(
                    self.client.upsert(collection_name=collection_name, data=[data])
                )
            except Exception as e:
                logger.error(f"Error adding document id: {data['id']} error: {e}")
                raise e
        return res

    def search(self, collection_name: str, bot_id: str, version: str, query: str, top_k: int = 4) -> List[RetrieverResult]:
        logger.info(
            f"Retreiver search for query: {query} on collection {collection_name} for bot_id: {bot_id} version: {version}"
        )
        
        partition_key = self.get_bot_uid(bot_id, version)
        query_embedding = embed(query)
        res = self.client.search(
            collection_name=collection_name,
            data=[query_embedding],
            limit=top_k,
            filter=f"bot_uid == '{partition_key}'",
            output_fields=["qa_doc_id", "chunk_id", "qa_doc_type", "metadata", "text"],
        )

        ret_results: List[RetrieverResult] = []
        for r in res[0]:
            logger.info(f"Milvus search result: {r}")
            qa_doc_id = r["entity"]["qa_doc_id"]
            chunk_id = r["entity"]["chunk_id"]
            text = r["entity"]["text"]
            logger.info(f"Retrieved qa_doc_id: {qa_doc_id} chunk_id: {chunk_id}")

            ret_results.append(RetrieverResult(
                qa_doc_id=qa_doc_id,
                qa_doc_type=RetrieverDocumentType(r["entity"]["qa_doc_type"]),
                text=text,
                metadata=r["entity"]["metadata"],
                distance=r["distance"],
                start_chunk_idx=chunk_id,
                end_chunk_idx=chunk_id,
            ))

        return ret_results

    ## TODO: get num_tokens for functions inside milvus_retriever.py and retriever_document.py (with classmethod RetrieverDocument.faq_retreiver_doc); influence token migrations
    def get_qa_docs(
        self, collection_name: str, bot_id: str, version: str, qa_doc_type: RetrieverDocumentType
    ) -> List[RetrieverDocument]:
        connections.connect(
            uri=self.uri,
            token=self.token,
        )
        collection = Collection(collection_name)

        partition_key = self.get_bot_uid(bot_id, version)
        iterator = collection.query_iterator(
            batch_size=1000,
            expr=f"qa_doc_type=='{qa_doc_type.value}' and bot_uid=='{partition_key}'",
            output_fields=[
                "id",
                "qa_doc_id",
                "chunk_id",
                "qa_doc_type",
                "text",
                "metadata",
                "bot_uid",
                "timestamp",
            ],
        )

        qa_docs = []
        qa_doc_map = defaultdict(list)
        while True:
            result = iterator.next()
            if len(result) == 0:
                iterator.close()
                break

            for r in result:
                if qa_doc_type == RetrieverDocumentType.FAQ:
                    qa_doc = RetrieverDocument.faq_retreiver_doc(
                        id=r["id"], text=r["text"], metadata=r["metadata"], bot_uid=r["bot_uid"], timestamp=r["timestamp"]
                    )
                    qa_docs.append(qa_doc)
                else:
                    qa_doc_map[r["qa_doc_id"]].append(r)
                    qa_doc = RetrieverDocument(
                        r["id"],
                        r["qa_doc_id"],
                        r["chunk_id"],
                        qa_doc_type,
                        r["text"],
                        r["metadata"],
                        is_chunked=True,
                        bot_uid=r["bot_uid"],
                        embedding=None,
                        timestamp=r["timestamp"],
                    )

        if qa_doc_type != RetrieverDocumentType.FAQ:
            for qa_doc_id, docs in qa_doc_map.items():
                sorted_docs = sorted(docs, key=lambda x: x["chunk_id"])
                txt = "".join([d["text"] for d in sorted_docs])
                ret_doc = RetrieverDocument.unchunked_retreiver_doc(
                    qa_doc_id,
                    qa_doc_type,
                    txt,
                    sorted_docs[0]["metadata"],
                    sorted_docs[0]["bot_uid"],
                    sorted_docs[0]["timestamp"],
                )
                qa_docs.append(ret_doc)

        return qa_docs

    def get_qa_doc(
        self, collection_name: str, qa_doc_id: str
    ) -> RetrieverDocument:

        logger.info(f"Getting qa doc with id {qa_doc_id} from collection {collection_name}")
        res = self.client.query(
            collection_name=collection_name,
            filter=f"qa_doc_id=='{qa_doc_id}'",
            output_fields=[
                "qa_doc_id",
                "chunk_id",
                "qa_doc_type",
                "text",
                "metadata",
                "bot_uid",
                "timestamp",
            ],
        )

        if len(res) == 0:
            return None

        sorted_res = sorted(res, key=lambda x: x["chunk_id"])

        if sorted_res[0]["qa_doc_type"] == RetrieverDocumentType.FAQ.value:
            return RetrieverDocument.faq_retreiver_doc(
                id=sorted_res[0]["qa_doc_id"],
                text=sorted_res[0]["text"],
                metadata=sorted_res[0]["metadata"],
                bot_uid=sorted_res[0]["bot_uid"],
                timestamp=sorted_res[0]["timestamp"],
            )
        else:
            txt = "".join([d["text"] for d in sorted_res])
            return RetrieverDocument.unchunked_retreiver_doc(
                sorted_res[0]["qa_doc_id"],
                RetrieverDocumentType(sorted_res[0]["qa_doc_type"]),
                txt,
                sorted_res[0]["metadata"],
                sorted_res[0]["bot_uid"],
                sorted_res[0]["timestamp"],
            )

    def get_qa_doc_ids(
        self, collection_name: str, bot_id: str, version: str, qa_doc_type: RetrieverDocumentType
    ) -> List[dict]:
        logger.info(f"Getting all qa_doc_ids from collection '{collection_name}' for bot_id: {bot_id}, version: {version}")
        partition_key = self.get_bot_uid(bot_id, version)
        connections.connect(
            uri=self.uri,
            token=self.token,
        )
        collection = Collection(collection_name)

        iterator = collection.query_iterator(
            batch_size=1000,
            expr=f"qa_doc_type == '{qa_doc_type.value}' and bot_uid == '{partition_key}'",
            output_fields=["id", "qa_doc_id"],
        )

        qa_doc_ids = set()
        while True:
            result = iterator.next()
            if len(result) == 0:
                iterator.close()
                break

            for r in result:
                qa_doc_ids.add(r["qa_doc_id"])

        return list(qa_doc_ids)

    def has_collection(self, collection_name: str):
        return self.client.has_collection(collection_name)

    def load_collection(self, collection_name: str):
        if self.client.has_collection(collection_name):
            self.client.load_collection(collection_name)
            return
        else:
            raise ValueError(f"Milvus Collection {collection_name} does not exist")

    def release_collection(self, collection_name: str):
        return self.client.release_collection(collection_name)
    
    def drop_collection(self, collection_name: str):
        return self.client.drop_collection(collection_name)
    
    def get_all_vectors(self, collection_name: str):
        connections.connect(
            uri=self.uri,
            token=self.token,
        )
        collection = Collection(collection_name)

        iterator = collection.query_iterator(
            batch_size=16000,
            output_fields=[
                'id',
                'qa_doc_id',
                "chunk_id",
                "qa_doc_type",
                "num_tokens",
                "metadata", 
                "text",
                "embedding",
                "timestamp",
            ],
        )

        vectors = []
        count = 0
        while True:
            result = iterator.next()
            if len(result) == 0:
                iterator.close()
                break

            for r in result:
                vectors.append(r)
                count += 1
        
        logger.info(f"collection {collection_name} Total vectors: {count}")
        return vectors

    def add_vectors_parallel(
        self, collection_name: str, bot_id: str, version: str, vectors: List[dict], upsert: bool = False
    ):
        logger.info(
            f"Adding {len(vectors)} vector db documents to institution {collection_name} for bot_id: {bot_id} version: {version}"
        )
        if not self.client.has_collection(collection_name):
            logger.info(f"No colelction found hence creating collection: {collection_name}")
            self.create_collection_with_partition_key(collection_name)

        vectors_to_insert = []

        if not upsert:
            # check if the document already exists in the collection
            for vec in vectors:
                res = self.client.query(
                    collection_name=collection_name, 
                    ids=vec["id"], 
                )
                if len(res) == 0:
                    vectors_to_insert.append(vec)
            logger.info(f"New vectors to insert: {len(vectors_to_insert)}")
        else:
            vectors_to_insert = vectors
        
        for vec in vectors_to_insert:
            vec["bot_uid"] = self.get_bot_uid(bot_id, version)

        res = []
        # process 100 documents at a time
        count = 0
        for i in range(0, len(vectors_to_insert), 100):
            batch_vectors = vectors_to_insert[i:i+100]

            res.extend(
                self.client.upsert(collection_name=collection_name, data=batch_vectors)
            )
            count += len(batch_vectors)
            logger.info(f"{collection_name}: Added {count}/{len(vectors_to_insert)} docs")
        return res

    def is_collection_loaded(self, collection_name: str):
        state = self.client.get_load_state(collection_name)
        print("loaded state: ", state)
        if state["state"].__str__() == "Loaded":
            return True
        else:
            return False
        
    def delete_vectors_by_partition_key(self, collection_name: str, bot_id: str, version: str):
        partition_key = self.get_bot_uid(bot_id, version)
        res = self.client.delete(collection_name=collection_name, filter=f"bot_uid=='{partition_key}'")
        logger.info(f"Deleted {len(res)} vectors from collection {collection_name} for bot_id: {bot_id} version: {version}")

        # check if the collection is empty
        res = self.client.query(collection_name=collection_name, output_fields=["count(*)"])
        if res[0]["count(*)"] == 0:
            logger.info(f"Collection {collection_name} is empty.")

        return res

    def get_vector_count_for_bot(self, collection: str, bot_id: str, version: str):
        res = self.client.query(
            collection_name=collection,
            filter=f"bot_uid=='{bot_id}__{version}'"
            )
        return len(res)
    
    # def get_token_count_for_bot(self, collection_name: str, bot_id: str, version: str):
    #     logger.info(f"Counting tokens in collection {collection_name} for bot_id: {bot_id}, version: {version}")
    #     partition_key = self.get_bot_uid(bot_id, version)
    #     res = self.client.query(
    #         collection_name=collection_name,
    #         filter=f"bot_uid=='{partition_key}'",
    #         output_fields=["num_tokens"],
    #     )
    #     return sum([r.get("num_tokens", 0) for r in res])
    
    def get_collection_size(self, collection_name: str):
        # real time vector count for the collection
        return self.client.query(collection_name=collection_name, output_fields=["count(*)"])[0]["count(*)"]
    
    def migrate_vectors(self, old_collection_name: str, bot_id: str, version: str, new_collection_name: str):
        partition_key = self.get_bot_uid(bot_id, version)
        connections.connect(
            uri=self.uri,
            token=self.token,
        )
        collection = Collection(old_collection_name)

        iterator = collection.query_iterator(
            batch_size=16000,
            expr=f"bot_uid=='{partition_key}'",
            output_fields=[
                'id',
                'bot_uid',
                'qa_doc_id',
                "chunk_id",
                "num_tokens",
                "qa_doc_type",
                "metadata", 
                "text",
                "embedding",
                "timestamp",
            ],
        )

        vectors = []
        count = 0
        while True:
            result = iterator.next()
            if len(result) == 0:
                iterator.close()
                break

            for r in result:
                vectors.append(r)
                count += 1

        logger.info(f"migrating {count} vectors for bot {bot_id} version {version}")
        
        # add vectors to new collection
        if not self.has_collection(new_collection_name):
            logger.info(f"No collection found hence creating collection: {new_collection_name}")
            self.create_collection_with_partition_key(new_collection_name)
        self.add_vectors_parallel(new_collection_name, bot_id, version, vectors)

        # delete vectors from old collection
        self.delete_vectors_by_partition_key(old_collection_name, bot_id, version)
        
        logger.info(f"moved {count} vectors from {old_collection_name} to {new_collection_name}")
        return count

    def list_collections(self):
        return self.client.list_collections()


class MilvusRetrieverExecutor:
    def __init__(self, bot_config):
        self.bot_config = bot_config
        self.llm = ChatOpenAI(model=MODEL["model_type_or_path"], timeout=30000)

    def generate_thought(self, retriever_results: List[RetrieverResult]) -> str:
        # post process list of documents into str
        retrieved_str = ""
        for doc in retriever_results:
            if doc.metadata.get("title"):
                retrieved_str += "title: " + doc.metadata["title"] + "\n"
            if doc.metadata.get("source"):
                retrieved_str += "source: " + doc.metadata["source"] + "\n"
            retrieved_str += "content: " + doc.text + "\n\n"
        return retrieved_str

    def _gaussian_similarity(self, distance, sigma=0.5):
        similarity = np.exp(-(distance**2) / (2 * sigma**2)) * 100
        return round(float(similarity), 2)

    def postprocess(self, retriever_results: List[RetrieverResult]):
        retriever_returns = []
        for doc in retriever_results:
            confidence_score = self._gaussian_similarity(doc.distance)
            item = {
                "qa_doc_id": doc.qa_doc_id,
                "qa_doc_type": doc.qa_doc_type.value,
                "title": doc.metadata.get("title"),
                "content": doc.text,
                "source": doc.metadata.get("source"),
                "raw_score": round(float(doc.distance), 4),
                "confidence": confidence_score,
            }
            retriever_returns.append(item)
        return {"retriever": retriever_returns}

    def retrieve(self, chat_history_str):
        """Given a chat history, retrieve relevant information from the database."""
        st = time.time()
        prompts = load_prompts(self.bot_config)
        contextualize_q_prompt = PromptTemplate.from_template(
            prompts.get("retrieve_contextualize_q_prompt", "")
        )
        ret_input_chain = contextualize_q_prompt | self.llm | StrOutputParser()
        ret_input = ret_input_chain.invoke({"chat_history": chat_history_str})
        rit = time.time() - st

        ret_results: List[RetrieverResult] = []
        st = time.time()
        milvus_db = mysql_pool.fetchone("SELECT collection_name FROM qa_bot WHERE id=%s AND version=%s", (self.bot_config.bot_id, self.bot_config.version))
        with MilvusRetriever() as retriever:
            ret_results = retriever.search(milvus_db["collection_name"], self.bot_config.bot_id, self.bot_config.version, ret_input)
        rt = time.time() - st
        logger.info(f"MilvusRetriever search took {rt} seconds")
        retriever_params = self.postprocess(ret_results)
        retriever_params["timing"] = {"retriever_input": rit, "retriever_search": rt}
        thought = self.generate_thought(ret_results)
        return thought, retriever_params


import os
import logging
import json
import uuid
import ast
import inspect

from arklex.utils.graph_state import MessageState, StatusEnum, Slot
from arklex.orchestrator.NLU.nlu import SlotFilling
from arklex.utils.utils import format_chat_history


logger = logging.getLogger(__name__)

    
def register_tool(desc, slots=[], outputs=[], isComplete=lambda x: True, isResponse=False):
    current_file_dir = os.path.dirname(__file__)
    def inner(func):
        file_path = inspect.getfile(func)
        relative_path = os.path.relpath(file_path, current_file_dir)
        # reformat the relative path to replace / and \\ with -, and remove .py, because the function calling in openai only allow the function name match the patter the pattern '^[a-zA-Z0-9_-]+$'
        # different file paths format in Windows and linux systems
        relative_path = relative_path.replace("/", "-").replace("\\", "-").replace(".py", "")
        key = f"{relative_path}-{func.__name__}"
        tool = lambda : Tool(func, key, desc, slots, outputs, isComplete, isResponse)
        return tool
    return inner

class Tool:
    def __init__(self, func, name, description, slots, outputs, isComplete, isResponse):
        self.func = func
        self.name = name
        self.description = description
        self.output = outputs
        self.slotfillapi: SlotFilling = None
        self.info = self.get_info(slots)
        self.slots = self._format_slots(slots)
        self.isComplete = isComplete
        self.isResponse = isResponse

    def _format_slots(self, slots):
        format_slots = []
        for slot in slots:
            format_slots.append(Slot(
                name=slot["name"], 
                type=slot["type"], 
                value="", 
                enum=slot.get("enum", []),
                description=slot["description"], 
                prompt=slot["prompt"], 
                required=slot.get("required", False),
                verified=slot.get("verified", False)
            ))
        return format_slots
    
    @staticmethod
    def format_slots(slots):
        format_slots = []
        for slot in slots:
            format_slots.append({
                "name": slot["name"],
                "value": slot["value"],
                "type": slot.get("type", "string"),
                "enum": slot.get("enum", []),
                "description": slot.get("description", ""),
                "prompt": slot.get("prompt", ""),
                "required": slot.get("required", False),
                "verified": slot.get("verified", False)
            })
        return format_slots

    def get_info(self, slots):
        self.properties = {}
        for slot in slots:
            self.properties[slot["name"]] = {k: v for k, v in slot.items() if k in ["type", "description", "prompt", "items"]}
        required = [slot["name"] for slot in slots if slot.get("required", False)]
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": {
                    "type": "object",
                    "properties": self.properties,
                    "required": required,
                },
            },
        }
        
    def init_slotfilling(self, slotfillapi: SlotFilling):
        self.slotfillapi = slotfillapi

    def _init_slots(self, state: MessageState):
        default_slots = state["slots"].get("default_slots", [])
        logger.info(f'Default slots are: {default_slots}')
        if not default_slots:
            return
        response = {}
        for default_slot in default_slots:
            response[default_slot.name] = default_slot.value
            for slot in self.slots:
                if slot.name == default_slot.name and default_slot.value:
                    slot.value = default_slot.value
                    slot.verified = True
        state["trajectory"].append({
            "role": "tool",
            "tool_call_id": str(uuid.uuid4()),
            "name": "default_slots",
            "content": json.dumps(response)
        })
        logger.info(f'Slots after initialization are: {self.slots}')

    def _skip_tool(self, state: MessageState):
        default_slots = state["slots"].get("default_slots", [])
        assigned_slot = False
        for default_slot in default_slots:
            for output in self.output:
                if output["name"] == default_slot.name and default_slot.value:
                    output["value"] = default_slot.value
                    assigned_slot = True
        if not assigned_slot:
            return False
        if all([output.get("value") for output in self.output if output.get("required", False)]):
            state["status"] = StatusEnum.COMPLETE.value
            call_id = str(uuid.uuid4())
            state["trajectory"].append({
                "role": "tool",
                "tool_call_id": call_id,
                "name": self.name,
                "content": json.dumps(self.output)
            })
            return True
        return False
        
    def _execute(self, state: MessageState, **fixed_args):
        # if this tool has been called before, then load the previous slots status
        if state["slots"].get(self.name):
            self.slots = state["slots"][self.name]
        else:
            state["slots"][self.name] = self.slots
        # init slot values saved in default slots
        self._init_slots(state)
        response = "error"
        # skip the tool if output slot is filled
        if self._skip_tool(state):
            logger.info("Tool is skipped because output slot is already filled")
            response = json.dumps(self.output)
            if self.isResponse:
                logger.info("Tool output is stored in response instead of message flow")
                state["response"] = response
            else:
                state["message_flow"] = state["message_flow"] + f"Context from {self.name}: {response}\n\n"
            return state
        max_tries = 3
        while max_tries > 0 and "error" in response:
            chat_history_str = format_chat_history(state["trajectory"])
            slots : list[Slot] = self.slotfillapi.execute(self.slots, chat_history_str, state["metadata"])
            logger.info(f'{slots=}')
            if not all([slot.value and slot.verified for slot in slots if slot.required]):
                for slot in slots:
                    # if there is extracted slots values but haven't been verified
                    if slot.value and not slot.verified:
                        # check whether it verified or not
                        verification_needed, thought = self.slotfillapi.verify_needed(slot, chat_history_str, state["metadata"])
                        if verification_needed:
                            response = slot.prompt + "The reason is: " + thought
                            break
                        else:
                            slot.verified = True
                    # if there is no extracted slots values, then should prompt the user to fill the slot
                    if not slot.value:
                        response = slot.prompt
                        break
                
                state["status"] = StatusEnum.INCOMPLETE.value
                
            # if slot.value is not empty for all slots, and all the slots has been verified, then execute the function
            if all([slot.value and slot.verified for slot in slots if slot.required]):
                logger.info("all slots filled")
                for slot in slots:
                    if slot.type in ["list", "dict", "array"]:
                        if not isinstance(slot.value, list):
                            try:
                                # Try to parse as JSON first
                                slot.value = json.loads(slot.value)
                            except json.JSONDecodeError:
                                # If JSON decoding fails, fallback to evaluating Python-like literals
                                try:
                                    slot.value = ast.literal_eval(slot.value)
                                except (ValueError, SyntaxError):
                                    raise ValueError(f"Unable to parse slot value: {slot.value}")
                kwargs = {slot.name: slot.value for slot in slots}
                combined_kwargs = {**kwargs, **fixed_args}
                response = self.func(**combined_kwargs)
                logger.info(f"Tool {self.name} response: {response}")
                call_id = str(uuid.uuid4())
                state["trajectory"].append({
                    'content': None, 
                    'role': 'assistant', 
                    'tool_calls': [
                        {
                            'function': {
                                'arguments': json.dumps(kwargs), 
                                'name': self.name
                            }, 
                            'id': call_id, 
                            'type': 'function'
                        }
                    ], 
                    'function_call': None
                })
                state["trajectory"].append({
                    "role": "tool",
                    "tool_call_id": call_id,
                    "name": self.name,
                    "content": response
                })
                if "error" in response:
                    max_tries -= 1
                    continue
                state["status"] = StatusEnum.COMPLETE.value if self.isComplete(response) else StatusEnum.INCOMPLETE.value

        if self.isResponse:
            logger.info("Tool output is stored in response instead of message flow")
            state["response"] = response
        else:
            state["message_flow"] = state["message_flow"] + f"Context from {self.name}: {response}\n\n"
        state["slots"][self.name] = slots
        return state

    def execute(self, state: MessageState, **fixed_args):
        state = self._execute(state, **fixed_args)
        ## postprocess if any
        ## Currently, the value of the tool is stored and returned in state["message_flow"]
        return state
    
    def __str__(self):
        return f"{self.__class__.__name__}"

    def __repr__(self):
        return f"{self.__class__.__name__}"
    


import requests
from arklex.env.tools.shopify.auth_utils import get_access_token

SHOPIFY_ADMIN_AUTH_ERROR = "error: missing some or all required shopify admin authentication parameters: shop_url, api_version, admin_token. Please set up 'fixed_args' in the config file. For example, {'name': <unique name of the tool>, 'fixed_args': {'admin_token': <shopify_access_token>, 'shop_url': <shopify_shop_url>, 'api_version': <Shopify API version>}}"
SHOPIFY_STOREFRONT_AUTH_ERROR = "error: missing some or all required shopify storefront authentication parameters: shop_url, api_version, storefront_token. Please set up 'fixed_args' in the config file. For example, {'name': <unique name of the tool>, 'fixed_args': {'storefront_token': <shopify_access_token>, 'shop_url': <shopify_shop_url>, 'api_version': <Shopify API version>}}"

def authorify_admin(kwargs):
    auth = {
        "value": {
            "domain": None,
            "version": None,
            "token": None
        },
        "error": None
    }
    auth["value"]["domain"] = kwargs.get("shop_url")
    auth["value"]["version"] = kwargs.get("api_version")
    auth["value"]["token"] = kwargs.get("admin_token")
    
    if not auth["value"]["domain"] or not auth["value"]["version"] or not auth["value"]["token"]:
        auth["error"] = SHOPIFY_ADMIN_AUTH_ERROR
        return auth
    
    return auth


def authorify_storefront(kwargs):
    auth = {
        "value": {
            "domain": None,
            "version": None,
            "token": None
        },
        "error": None,
        "storefront_token": None,
        "storefront_url": None
    }
    auth["value"]["domain"] = kwargs.get("shop_url")
    auth["value"]["version"] = kwargs.get("api_version")
    auth["value"]["token"] = kwargs.get("storefront_token")
    
    if not auth["value"]["domain"] or not auth["value"]["version"] or not auth["value"]["token"]:
        auth["error"] = SHOPIFY_STOREFRONT_AUTH_ERROR
        return auth
    
    auth["storefront_token"] = auth["value"]["token"]
    auth["storefront_url"] = f"{auth['value']['domain']}/api/{auth['value']['version']}/graphql.json"
    return auth


def make_query(url, query, variables, headers):
    """
    Make query response
    """
    request = requests.post(url, json={'query': query, 'variables': variables}, headers=headers)
    if request.status_code == 200:
        return request.json()
    else:
        raise Exception("Query failed to run by returning code of {}. {}".format(request.status_code, query))



from typing import Any, Dict
import json

import shopify

from arklex.env.tools.tools import register_tool
from arklex.env.tools.shopify.utils import authorify_admin
from arklex.env.tools.shopify.utils_slots import ShopifySlots, ShopifyOutputs

description = "Find user id by email. If the user is not found, the function will return an error message."
slots = [
    ShopifySlots.USER_EMAIL
]
outputs = [
    ShopifyOutputs.USER_ID
]

USER_NOT_FOUND_ERROR = "error: user not found"
MULTIPLE_USERS_SAME_EMAIL_ERROR = "error: there are multiple users with the same email"
errors = [
    USER_NOT_FOUND_ERROR,
    MULTIPLE_USERS_SAME_EMAIL_ERROR
]

@register_tool(description, slots, outputs, lambda x: x not in errors)
def find_user_id_by_email(user_email: str, **kwargs) -> str:
    auth = authorify_admin(kwargs)
    if auth["error"]:
        return auth["error"]
    
    user_id = ""
    
    try:
        with shopify.Session.temp(**auth["value"]):
            response = shopify.GraphQL().execute(f"""
                {{
                    customers (first: 10, query: "email:{user_email}") {{
                        edges {{
                            node {{
                                id
                            }}
                        }}
                    }}
                }}
                """)
        nodes = json.loads(response)["data"]["customers"]["edges"]
        if len(nodes) == 1:
            user_id = nodes[0]["node"]["id"]
            return user_id
        else:
            return MULTIPLE_USERS_SAME_EMAIL_ERROR
    except Exception as e:
        return USER_NOT_FOUND_ERROR


import logging

from langchain.prompts import PromptTemplate
from langchain_openai import ChatOpenAI
from langchain_core.output_parsers import StrOutputParser

from arklex.env.workers.worker import BaseWorker, register_worker
from arklex.env.prompts import load_prompts
from arklex.utils.utils import chunk_string
from arklex.utils.graph_state import MessageState
from arklex.utils.model_config import MODEL
from arklex.utils.model_provider_config import PROVIDER_MAP


logger = logging.getLogger(__name__)


@register_worker
class DefaultWorker(BaseWorker):

    description = "Default worker decided by chat records if there is no specific worker for the user's query"

    def __init__(self):
        super().__init__()
        self.llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
            model=MODEL["model_type_or_path"], timeout=30000
        )

    def _choose_worker(self, state: MessageState, limit=2):
        user_message = state['user_message']
        task = state["orchestrator_message"].attribute.get("task", "")
        self.available_workers = {id: resource for id, resource in state["metadata"]["worker"].items() if resource["name"] != "DefaultWorker"}
        self.name2id = {resource["name"]: id for id, resource in self.available_workers.items()}
        workers_info = "\n".join([f"{resource['name']}: {resource['description']}" for _, resource in self.available_workers.items()])
        workers_name = ", ".join(self.name2id.keys())

        prompts = load_prompts(state["bot_config"])
        prompt = PromptTemplate.from_template(prompts["choose_worker_prompt"])
        input_prompt = prompt.invoke({"message": user_message.message, "formatted_chat": user_message.history, "task": task, "workers_info": workers_info, "workers_name": workers_name})
        chunked_prompt = chunk_string(input_prompt.text, tokenizer=MODEL["tokenizer"], max_length=MODEL["context"])
        final_chain = self.llm | StrOutputParser()
        while limit > 0:
            answer = final_chain.invoke(chunked_prompt)
            for worker_name in self.name2id.keys():
                if worker_name in answer:
                    logger.info(f"Chosen worker name for the default worker: {worker_name}")
                    worker_id = self.name2id[worker_name]
                    return worker_id
            limit -= 1
        logger.info("Worker chosen failed for the default worker.")
        return ""
    
    def execute(self, msg_state: MessageState):
        worker_id = self._choose_worker(msg_state)
        if worker_id:
            worker = self.available_workers[worker_id]["execute"]()
            result = worker.execute(msg_state)
        else:
            result = msg_state
        return result



from __future__ import annotations

import argparse
from typing import Any, TypeVar

from pydantic import BaseModel

from benchmark.tau_bench.model_utils.api._model_methods import MODEL_METHODS
from benchmark.tau_bench.model_utils.api.cache import cache_call_w_dedup
from benchmark.tau_bench.model_utils.api.datapoint import (
    BinaryClassifyDatapoint,
    ClassifyDatapoint,
    Datapoint,
    GenerateDatapoint,
    ParseDatapoint,
    ParseForceDatapoint,
    ScoreDatapoint,
)
from benchmark.tau_bench.model_utils.api.logging import log_call
from benchmark.tau_bench.model_utils.api.router import RequestRouter, default_request_router
from benchmark.tau_bench.model_utils.api.sample import (
    EnsembleSamplingStrategy,
    MajoritySamplingStrategy,
    SamplingStrategy,
    get_default_sampling_strategy,
)
from benchmark.tau_bench.model_utils.api.types import PartialObj
from benchmark.tau_bench.model_utils.model.general_model import GeneralModel
from benchmark.tau_bench.model_utils.model.model import (
    AnyModel,
    BinaryClassifyModel,
    ClassifyModel,
    GenerateModel,
    ParseForceModel,
    ParseModel,
    ScoreModel,
)

T = TypeVar("T", bound=BaseModel)


class API(object):
    wrappers_for_main_methods = [log_call, cache_call_w_dedup]

    def __init__(
        self,
        parse_models: list[ParseModel],
        generate_models: list[GenerateModel],
        parse_force_models: list[ParseForceModel],
        score_models: list[ScoreModel],
        classify_models: list[ClassifyModel],
        binary_classify_models: list[BinaryClassifyModel] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        log_file: str | None = None,
    ) -> None:
        if sampling_strategy is None:
            sampling_strategy = get_default_sampling_strategy()
        if request_router is None:
            request_router = default_request_router()
        self.sampling_strategy = sampling_strategy
        self.request_router = request_router
        self._log_file = log_file
        self.binary_classify_models = binary_classify_models
        self.classify_models = classify_models
        self.parse_models = parse_models
        self.generate_models = generate_models
        self.parse_force_models = parse_force_models
        self.score_models = score_models

        self.__init_subclass__()

        self.__init_subclass__()

    def __init_subclass__(cls):
        for method_name in MODEL_METHODS:
            if hasattr(cls, method_name):
                method = getattr(cls, method_name)
                for wrapper in cls.wrappers_for_main_methods:
                    method = wrapper(method)
                setattr(cls, method_name, method)

    @classmethod
    def from_general_model(
        cls,
        model: GeneralModel,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        log_file: str | None = None,
    ) -> "API":
        return cls(
            binary_classify_models=[model],
            classify_models=[model],
            parse_models=[model],
            generate_models=[model],
            parse_force_models=[model],
            score_models=[model],
            log_file=log_file,
            sampling_strategy=sampling_strategy,
            request_router=request_router,
        )

    @classmethod
    def from_general_models(
        cls,
        models: list[GeneralModel],
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        log_file: str | None = None,
    ) -> "API":
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        return cls(
            binary_classify_models=models,
            classify_models=models,
            parse_models=models,
            generate_models=models,
            parse_force_models=models,
            score_models=models,
            log_file=log_file,
            sampling_strategy=sampling_strategy,
            request_router=request_router,
        )

    def set_default_binary_classify_models(self, models: list[BinaryClassifyModel]) -> None:
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.binary_classify_models = models

    def set_default_classify_models(self, models: list[BinaryClassifyModel]) -> None:
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.classify_models = models

    def set_default_parse_models(self, models: list[ParseModel]) -> None:
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.parse_models = models

    def set_default_generate_models(self, models: list[GenerateModel]) -> None:
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.generate_models = models

    def set_default_parse_force_models(self, models: list[ParseForceModel]) -> None:
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.parse_force_models = models

    def set_default_score_models(self, models: list[ScoreModel]) -> None:
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.score_models = models

    def set_default_sampling_strategy(self, sampling_strategy: SamplingStrategy) -> None:
        self.sampling_strategy = sampling_strategy

    def set_default_request_router(self, request_router: RequestRouter) -> None:
        self.request_router = request_router

    def _run_with_sampling_strategy(
        self,
        models: list[AnyModel],
        datapoint: Datapoint,
        sampling_strategy: SamplingStrategy,
    ) -> T:
        assert len(models) > 0

        def _run_datapoint(model: AnyModel, temp: float | None = None) -> T:
            if isinstance(datapoint, ClassifyDatapoint):
                return model.classify(
                    instruction=datapoint.instruction,
                    text=datapoint.text,
                    options=datapoint.options,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            elif isinstance(datapoint, BinaryClassifyDatapoint):
                return model.binary_classify(
                    instruction=datapoint.instruction,
                    text=datapoint.text,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            elif isinstance(datapoint, ParseForceDatapoint):
                return model.parse_force(
                    instruction=datapoint.instruction,
                    typ=datapoint.typ,
                    text=datapoint.text,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            elif isinstance(datapoint, GenerateDatapoint):
                return model.generate(
                    instruction=datapoint.instruction,
                    text=datapoint.text,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            elif isinstance(datapoint, ParseDatapoint):
                return model.parse(
                    text=datapoint.text,
                    typ=datapoint.typ,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            elif isinstance(datapoint, ScoreDatapoint):
                return model.score(
                    instruction=datapoint.instruction,
                    text=datapoint.text,
                    min=datapoint.min,
                    max=datapoint.max,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            else:
                raise ValueError(f"Unknown datapoint type: {type(datapoint)}")

        if isinstance(sampling_strategy, EnsembleSamplingStrategy):
            return sampling_strategy.execute(
                [lambda x=model: _run_datapoint(x, 0.0) for model in models]
            )
        return sampling_strategy.execute(
            lambda: _run_datapoint(
                models[0], 0.2 if isinstance(sampling_strategy, MajoritySamplingStrategy) else None
            )
        )

    def _api_call(
        self, models: list[AnyModel], datapoint: Datapoint, sampling_strategy: SamplingStrategy
    ) -> T:
        if isinstance(sampling_strategy, EnsembleSamplingStrategy):
            return self._run_with_sampling_strategy(models, datapoint, sampling_strategy)
        model = self.request_router.route(dp=datapoint, available_models=models)
        return self._run_with_sampling_strategy(
            models=[model], datapoint=datapoint, sampling_strategy=sampling_strategy
        )

    def classify(
        self,
        instruction: str,
        text: str,
        options: list[str],
        examples: list[ClassifyDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[ClassifyModel] | None = None,
    ) -> int:
        if models is None:
            models = self.classify_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router

        return self._api_call(
            models=models,
            datapoint=ClassifyDatapoint(
                instruction=instruction, text=text, options=options, examples=examples
            ),
            sampling_strategy=sampling_strategy,
        )

    def binary_classify(
        self,
        instruction: str,
        text: str,
        examples: list[BinaryClassifyDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[BinaryClassifyModel] | None = None,
    ) -> bool:
        if models is None:
            models = (
                self.binary_classify_models
                if self.binary_classify_models is not None
                else self.classify_models
            )
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router

        return self._api_call(
            models=models,
            datapoint=BinaryClassifyDatapoint(
                instruction=instruction, text=text, examples=examples
            ),
            sampling_strategy=sampling_strategy,
        )

    def parse(
        self,
        text: str,
        typ: type[T] | dict[str, Any],
        examples: list[ParseDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[ParseModel] | None = None,
    ) -> T | PartialObj | dict[str, Any]:
        if models is None:
            models = self.parse_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router

        return self._api_call(
            models=models,
            datapoint=ParseDatapoint(text=text, typ=typ, examples=examples),
            sampling_strategy=sampling_strategy,
        )

    def generate(
        self,
        instruction: str,
        text: str,
        examples: list[GenerateDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[GenerateModel] | None = None,
    ) -> str:
        if models is None:
            models = self.generate_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router

        return self._api_call(
            models=models,
            datapoint=GenerateDatapoint(instruction=instruction, text=text, examples=examples),
            sampling_strategy=sampling_strategy,
        )

    def parse_force(
        self,
        instruction: str,
        typ: type[T] | dict[str, Any],
        text: str | None = None,
        examples: list[ParseForceDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[ParseForceModel] | None = None,
    ) -> T | dict[str, Any]:
        if models is None:
            models = self.parse_force_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router

        return self._api_call(
            models=models,
            datapoint=ParseForceDatapoint(
                instruction=instruction, typ=typ, text=text, examples=examples
            ),
            sampling_strategy=sampling_strategy,
        )

    def score(
        self,
        instruction: str,
        text: str,
        min: int,
        max: int,
        examples: list[ScoreDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[ScoreModel] | None = None,
    ) -> int:
        if models is None:
            models = self.score_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router

        return self._api_call(
            models=models,
            datapoint=ScoreDatapoint(
                instruction=instruction, text=text, min=min, max=max, examples=examples
            ),
            sampling_strategy=sampling_strategy,
        )


def default_api(
    log_file: str | None = None,
    sampling_strategy: SamplingStrategy | None = None,
    request_router: RequestRouter | None = None,
) -> API:
    from benchmark.tau_bench.model_utils.model.general_model import default_model

    model = default_model()
    return API(
        binary_classify_models=[model],
        classify_models=[model],
        parse_models=[model],
        generate_models=[model],
        parse_force_models=[model],
        score_models=[model],
        sampling_strategy=sampling_strategy,
        request_router=request_router,
        log_file=log_file,
    )

def default_api_from_args(args: argparse.Namespace) -> API:
    from benchmark.tau_bench.model_utils.model.general_model import model_factory
    model = model_factory(model_id=args.model, platform=args.platform, base_url=args.base_url)
    return API.from_general_model(model=model)


def default_quick_api(
    log_file: str | None = None,
    sampling_strategy: SamplingStrategy | None = None,
    request_router: RequestRouter | None = None,
) -> API:
    from benchmark.tau_bench.model_utils.model.general_model import default_quick_model

    model = default_quick_model()
    return API(
        binary_classify_models=[model],
        classify_models=[model],
        parse_models=[model],
        generate_models=[model],
        parse_force_models=[model],
        score_models=[model],
        sampling_strategy=sampling_strategy,
        request_router=request_router,
        log_file=log_file,
    )



# Copyright Sierra

from setuptools import find_packages, setup

setup(
    name="tau_bench",
    version="0.1.0",
    description="The Tau-Bench package",
    long_description=open("README.md").read(),
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "openai>=1.13.3",
        "mistralai>=0.4.0",
        "anthropic>=0.26.1",
        "google-generativeai>=0.5.4",
        "tenacity>=8.3.0",
        "termcolor>=2.4.0",
        "numpy>=1.26.4",
        "litellm>=1.41.0",
    ],
)
